/* -*- indent-tabs-mode: t; tab-width: 4; c-basic-offset: 4; -*- */
/*
   hashing_struct.c
   Projet GtkCmpHash

   (C) Copyright 2007 Olivier Delhomme
   e-mail : olivierdelhomme@gmail.com
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.           
*/
/**
 *  Gestion du chargement, de la mise en mmoire et des comparaisons
 *  des hashs
 *  Utilisation de la structure :
 *   - Cration de l'arbre/tronc : nouveau_tronc(niveau)
 *     niveau indique le nombre d'indirections (intressant  partir de 2)
 *   - Insertion dans l'arbre/tronc : insere_dans_tronc(tronc, file_hash)
 *     Insre file_hash dans le tronc (l o il faut)
 *   - Recherche : recherche_dans_tronc(tronc, file_hash)
 *     Recherche file_hash rapidement (en quelques comparaisons) avec 
 *     gestion des collisions
 */

#include "gtkcmphash.h"

static arbre_t *nouvel_arbre(guint niveau_max, guint niveau);
static guint cle_de_hashage(guchar c);
static guint int_value_of_hex(guchar c);
static GSList *insere_dans_liste(GSList *list, file_hash_t *file_hash);
static void insere_dans_arbre(arbre_t *arbre, guint niveau_max, file_hash_t *file_hash);
static GSList * recherche_dans_arbre(arbre_t *arbre, guint niveau_max, file_hash_t *file_hash);
static GSList *recherche_dans_liste(GSList *list, file_hash_t *file_hash);

/* Fonctions de la glib modifies */
static gint my_g_ascii_strcasecmp (const guchar *s1, const guchar *s2, guint len);
static guchar *my_g_strdup (const guchar *str);

static result_hash_t *copie_les_resultats(file_hash_t *file_hash, file_hash_t *result_hash);
static GSList *forme_la_liste_des_connus(GSList *found, GSList *result, file_hash_t *file_hash, options_t *opts);

static GSList *concatene_les_arbres(arbre_t *arbre, guint niveau_max);
static GSList *concatene_les_listes(arbre_t *arbre);
static GSList *transforme_arbre_en_liste(arbre_t *arbre, guint niveau_max);


/**
 *  Cre un nouvel arbre de profondeur "niveau_max",  partir de "niveau"
 */
static arbre_t *nouvel_arbre(guint niveau_max, guint niveau)
{
  arbre_t *arbre = NULL;
  guint i = 0;

  arbre = (arbre_t *) g_malloc0(sizeof(arbre_t));
  arbre->niveau = niveau;
  
  if (niveau >= niveau_max)
    { 
      for (i=0; i<16; i++)
	arbre->array[i] = NULL;
      return arbre;
    }
  else
    {
      for (i=0; i<16; i++)
	arbre->array[i] = nouvel_arbre(niveau_max, niveau+1);
      return arbre;
    }
}

/**
 *  Construit une structure vide de niveau "niveau"
 *  maximum 16 sous niveaux pour le md5 
 */
hash_t *nouveau_tronc(guint niveau_max)
{
  hash_t *tronc = NULL;

  tronc = (hash_t *) g_malloc0(sizeof(hash_t));

  tronc->arbre = nouvel_arbre(niveau_max, 1);
  tronc->nb_niveau = niveau_max;

  return tronc;
}

/**
 *  Cl de hashage ( partir d'un octet renvoi 16
 *  cas possibles)
 */
static guint cle_de_hashage(guchar c)
{
  return ((guint)(c >>4));
}

/**
 *  Retourne la valeur du caractre hxadcimal en dcimal
 *  0->0 ... 9->9, a->10, ... f->15 sinon 0
 */
static guint int_value_of_hex(guchar c)
{
  switch (c)
    {
    case '0':
      return 0;
      break;
    case '1':
      return 1;
      break;
    case '2':
      return 2;
      break;
    case '3':
      return 3;
      break;
    case '4':
      return 4;
      break;
    case '5':
      return 5;
      break;
    case '6':
      return 6;
      break;
    case '7':
      return 7;
      break;
    case '8':
      return 8;
      break;
    case '9':
      return 9;
      break;
    case 'a':
      return 10;
      break;
    case 'b':
      return 11;
      break;
    case 'c':
      return 12;
      break;
    case 'd':
      return 13;
      break;
    case 'e':
      return 14;
      break;
    case 'f':
      return 15;
      break;
    default:
      return 0;
      break;
    }
}

/**
 *  Insre le file_hash d'un fichier dans une liste
 *  cre ventuellement cette liste
 */
static GSList *insere_dans_liste(GSList *list, file_hash_t *file_hash)
{
  list = g_slist_prepend(list, file_hash);

  return list;
}

/**
 *  Insre un file_hash d'un fichier dans l'arbre (en fonction de son md5)
 *  TODO : laisser le choix  l'utilisateur du hash servant de comparaison
 */
static void insere_dans_arbre(arbre_t *arbre, guint niveau_max, file_hash_t *file_hash)
{
  guint numero_liste = 0;
  guchar *hash = file_hash->hash_md5;

  numero_liste = cle_de_hashage(hash[arbre->niveau-1]);

  if (arbre->niveau >= niveau_max)
    arbre->array[numero_liste] = insere_dans_liste(arbre->array[numero_liste], file_hash); 
  else  
    insere_dans_arbre(arbre->array[numero_liste], niveau_max, file_hash); 
}

/**
 *  Insre un file_hash d'un fichier dans la structure
 */
void insere_dans_tronc(hash_t *tronc, file_hash_t *file_hash)
{
  
  if (tronc != NULL)
    {
      if (tronc->arbre != NULL && tronc->nb_niveau > 0)
	{
	  insere_dans_arbre(tronc->arbre, tronc->nb_niveau, file_hash);
	}
    }
}

/**
 *  Insre toute une liste de file_hash dans la structure tronc
 */
void insere_la_liste_dans_tronc(hash_t *tronc, GSList *file_hash_list)
{
  file_hash_t *file_hash = NULL;

  while (file_hash_list != NULL)
    {
      file_hash = (file_hash_t *) file_hash_list->data;

      if (file_hash != NULL)
	insere_dans_tronc(tronc, file_hash);

      file_hash_list = g_slist_next(file_hash_list);
    }
}

/**
 *  Procdure copie de la glib, mais dans laquelle
 *  j'ai enlev les TOLOWER dont je n'ai absolument 
 *  pas besoin... donc gain de rapidit (c'est plus rapide que memcmp)
 */
static gint my_g_ascii_strcasecmp(const guchar *s1, const guchar *s2, guint len)
{

  while (len)
    {
      if ((gint)(guchar)(*s1) != (gint)(guchar)(*s2))
	return (((gint)(guchar) *s1) - ((gint)(guchar) *s2));
	
      s1++; 
      s2++; 
      len--;
    }

  return (((gint)(guchar) *s1) - ((gint)(guchar) *s2));
}

/**
 *  Procdure rcupre de la glib et transforme pour
 *  qu'elle ait en entre et en sortie le type guchar
 */
static guchar *my_g_strdup (const guchar *str)
{
  guchar *new_str;
  gsize length;

  if (str)
    {
		length = strlen((const char *) str) + 1;
      new_str = g_new(unsigned char, length);
      memcpy(new_str, str, length);
    }
  else
    new_str = NULL;

  return new_str;
}



/**
 *  Recherche dans la liste d'ventuels fichiers ayant un hash 
 *  identique  celui de file_hash (utilise le md5)
 */
static GSList *recherche_dans_liste(GSList *list, file_hash_t *file_hash)
{
   file_hash_t *tmp_hash = NULL;
   GSList *head = list;
   GSList *file_hash_list = NULL;

  while (list != NULL)
    {
      tmp_hash = list->data;
      
      if (tmp_hash != NULL)
	{
	  if (my_g_ascii_strcasecmp(file_hash->hash_md5, tmp_hash->hash_md5, tmp_hash->len_md5) == 0) 
	    file_hash_list = g_slist_prepend(file_hash_list, tmp_hash);	
	}

      list = list->next;
    }
  
  list = head;
  
  return file_hash_list;
}

/**
 *  Effectue la recherche dans un arbre
 */
static GSList *recherche_dans_arbre(arbre_t *arbre, guint niveau_max, file_hash_t *file_hash)
{
  guint numero_liste = 0;
  guchar *hash = file_hash->hash_md5;

  /* il faut utiliser la mme clef que celle de l'insertion */
  /* TODO : donner le choix de la clef  l'utilisateur      */
  numero_liste = cle_de_hashage(hash[arbre->niveau-1]);

  if (arbre->niveau >= niveau_max)
    return recherche_dans_liste(arbre->array[numero_liste], file_hash);
  else
    return recherche_dans_arbre(arbre->array[numero_liste], niveau_max, file_hash);
}

/**
 *  Recherche si le hash "hash" est prsent dans la structure de donnes
 *  la fonction retourne la liste des hash correspondant
 *  NULL dans les autres cas
 */ 
GSList *recherche_dans_tronc(hash_t *tronc, file_hash_t *file_hash)
{
  if (tronc != NULL && tronc->arbre != NULL && tronc->nb_niveau > 0)
    return recherche_dans_arbre(tronc->arbre, tronc->nb_niveau, file_hash);
  else
    return NULL;
}

/**
 *  Concatne les sous-listes
 */
static GSList *concatene_les_listes(arbre_t *arbre)
{
  guint i = 0;
  GSList *en_cours = NULL;

  for (i = 0; i < 16; i++)
    {
      if (arbre->array[i] != NULL)
	  en_cours = g_slist_concat(arbre->array[i], en_cours);
    }
 
 return en_cours;
}

/**
 *  Concatne le rsultat des diffrents sous-arbres
 */
static GSList *concatene_les_arbres(arbre_t *arbre, guint niveau_max)
{ 
  guint i = 0;
  GSList *en_cours = NULL;
  GSList *resultats = NULL;

  for (i = 0; i < 16; i++)
    {
      resultats = transforme_arbre_en_liste(arbre->array[i], niveau_max);
      if (resultats != NULL)
	  en_cours = g_slist_concat(resultats, en_cours);
    }

  return en_cours;
}

/**
 *  Remet l'arbre  plat
 */
static GSList *transforme_arbre_en_liste(arbre_t *arbre, guint niveau_max)
{
  if (arbre->niveau >= niveau_max)  
    return concatene_les_listes(arbre); /* ici c'est le niveau des listes */ 
  else
    return concatene_les_arbres(arbre, niveau_max);
}

/**
 *  Remet le tronc  plat
 */ 
GSList *transforme_tronc_en_liste(hash_t *tronc)
{
  if (tronc != NULL && tronc->arbre != NULL && tronc->nb_niveau > 0)
    return transforme_arbre_en_liste(tronc->arbre, tronc->nb_niveau);
  else
    return NULL;
}

/**
 *  Procdure qui transforme le hash b1f548... en son
 *  quivalent binaire; Ralise un g_malloc0 qu'on pourra
 *  librer avec un g_free par la suite.
 */
guchar *transforme_le_hash_de_hex_en_binaire(guchar *hash)
{
  guchar c = (guchar) 0;
  guint i = 0;
  guint j = 0;
  guint len = strlen((const char *)hash);
  guchar *aux = NULL;

  aux = (guchar *) g_malloc0(sizeof(guchar)*((len/2)+2));
  
  while (i < len)
    {
      c = 16*int_value_of_hex(hash[i]) + int_value_of_hex(hash[i+1]);
      aux[j] = c;
      j++;
      i +=2;
    }
  aux[j] = (guchar) 0;

  return aux;
}

/**
 *  Fonction qui transforme un hash sous forme binaire
 *  en un hash sous forme hexadcimale affichable "12f5cda5..."
 *  Ralise un g_malloc0 qu'on pourra librer par la suite avec
 *  un g_free
 */
guchar *transforme_le_hash_de_binaire_en_hex(guchar *hash, guint len)
{
  guint i = 0;
  guchar *aux = NULL;
  guchar *md5 = NULL;

  aux = (guchar *) g_malloc0(sizeof(gchar)*3);
  md5 = (guchar *) g_malloc0(sizeof(guchar)*((len*2)+2));

   for(i = 0; i < len; i++)  
    {
      sprintf((char *) aux, "%02x", hash[i]); 
      memcpy(md5+(i*2), aux, 2);
    }
   md5[i*2+1] = '\0';
   g_free(aux);

   return md5;
}

/**
 *  Dit si l'un des hash de file_hash est un hash de fichier vide
 */
gboolean is_file_hash_empty(file_hash_t *file_hash)
{
	guchar *md5_vide = transforme_le_hash_de_hex_en_binaire((guchar *) "d41d8cd98f00b204e9800998ecf8427e");
	guchar *sha1_vide = transforme_le_hash_de_hex_en_binaire((guchar *) "da39a3ee5e6b4b0d3255bfef95601890afd80709");
	guchar *ripemd_vide = transforme_le_hash_de_hex_en_binaire((guchar *) "9c1185a5c5e9fc54612808977ee8f548b2258d31");
  gboolean result = FALSE;

  if (file_hash != NULL)
    {
      result = my_g_ascii_strcasecmp(file_hash->hash_md5, md5_vide, file_hash->len_md5);

      if (result != TRUE)
	result = my_g_ascii_strcasecmp(file_hash->hash_sha1, sha1_vide, file_hash->len_sha1);

      if (result != TRUE)
	result = my_g_ascii_strcasecmp(file_hash->hash_ripemd, ripemd_vide, file_hash->len_ripemd);
    }

  g_free(md5_vide);
  g_free(sha1_vide);
  g_free(ripemd_vide);

  return result;
}

/**
 *  Fonction de libration d'un file_hash
 *  On ne supprime pas le hashset_name qui n'est pas inclus dans
 *  chaque structure (donc si on le libre en cours de route, 
 *  on le perd dfinitivement)
 */
void free_file_hash(file_hash_t *file_hash)
{
  if (file_hash != NULL)
	{
	  if (file_hash->hash_md5 != NULL)
	    g_free(file_hash->hash_md5);
	  if (file_hash->hash_sha1 != NULL)
	    g_free(file_hash->hash_sha1);
	  if (file_hash->hash_ripemd != NULL)
	    g_free(file_hash->hash_ripemd);
	  if (file_hash->filename != NULL)
	    g_free(file_hash->filename);
	}
}

/**
 *  Fonction de libration de la liste (libration totale)
 *  perte du hashset_name !
 */
void free_file_hash_list(GSList *file_hash_list)
{
  GSList *head = file_hash_list;
  file_hash_t *file_hash = NULL;

  while (file_hash_list != NULL)
    {
      file_hash = (file_hash_t *) file_hash_list->data;
      
      free_file_hash(file_hash);

      if (file_hash->hashset_name != NULL)
	    g_free(file_hash->hashset_name);
      
      file_hash_list = g_slist_next(file_hash_list);
    }
  
  g_slist_free(head);
}

/**
 *  Copie les informations adquates dans une nouvelle
 *  structure qui est retourne
 */
static result_hash_t *copie_les_resultats(file_hash_t *file_hash, file_hash_t *result_hash)
{
  result_hash_t *all_known_hash = NULL;
 
  all_known_hash = (result_hash_t *) g_malloc0(sizeof(result_hash_t));

  all_known_hash->hashset_name = g_strdup(result_hash->hashset_name);
  all_known_hash->hashset_file_filename = g_strdup(result_hash->filename);

  all_known_hash->filename = g_strdup(file_hash->filename);
  all_known_hash->hash_md5 = my_g_strdup(file_hash->hash_md5);
  all_known_hash->len_md5 = file_hash->len_md5;
  all_known_hash->hash_sha1 = my_g_strdup(file_hash->hash_sha1);
  all_known_hash->len_sha1 = file_hash->len_sha1;
  all_known_hash->hash_ripemd = my_g_strdup(file_hash->hash_ripemd);
  all_known_hash->len_ripemd = file_hash->len_ripemd;
  
  return all_known_hash;
}

/**
 *  Insre, en fonction des options, les fichiers connus dans la liste
 *  Attention la liste contient des lments de type result_hash_t *
 */
static GSList *forme_la_liste_des_connus(GSList *found, GSList *result, file_hash_t *file_hash, options_t *opts)
{
  file_hash_t *result_hash = NULL;
  result_hash_t *all_known_hash = NULL;

  if (opts->all_known == TRUE)
    {
      /* intgre tous les fichiers connus */
      while (result != NULL)
	{
	  result_hash = (file_hash_t *) result->data;
	 
	  all_known_hash = copie_les_resultats(file_hash, result_hash);
	 
	  found = g_slist_prepend(found, all_known_hash);

	  result = g_slist_next(result);
	  /* On ne libre pas de mmoire car tout est dans la liste pour une   */
	  /* utilisation ultrieure. La libration intervient dans la fonction */
	  /* free_file_hash_list(GSList *file_hash_list) plus haut             */
	}
    }
  else
    {
      result_hash = (file_hash_t *) result->data;

      all_known_hash = copie_les_resultats(file_hash, result_hash);
     
      found = g_slist_prepend(found, all_known_hash);
    }

  return found;
}

/**
 *  Fonction pour trouver les diffrences entre une liste de fichiers
 *  hachs et un hashset chargs en mmoire
 */
found_or_not_t *find_all_hashes_from_hashset(GSList *file_hash_list, hash_t *tronc, options_t *opts)
{
  file_hash_t *file_hash = NULL; /* pour parcourir la liste des fichiers hachs */
  GSList *result = NULL;  /* rsultat de la recherche dans l'arbre */
  found_or_not_t *dedans_ou_pas = NULL;

  dedans_ou_pas = (found_or_not_t *) g_malloc0(sizeof(found_or_not_t));
  dedans_ou_pas->found = NULL;
  dedans_ou_pas->not_found = NULL;
  
  while (file_hash_list != NULL)
    {
      file_hash = (file_hash_t *) file_hash_list->data;

      if (file_hash != NULL)
	{
	  /* retourne la liste de tous les fichiers qui ont le mme hash */
	  result = recherche_dans_tronc(tronc, file_hash); 

	  if (result != NULL)
	    dedans_ou_pas->found = forme_la_liste_des_connus(dedans_ou_pas->found, result, file_hash, opts);
	  else
	    dedans_ou_pas->not_found = g_slist_prepend(dedans_ou_pas->not_found, file_hash);

	  g_slist_free(result);
	}

      file_hash_list = g_slist_next(file_hash_list);
    }
  
  return dedans_ou_pas;
}
